#include <String.h>
#include <iostream.h>
#include "Objects.h"

#include "Expr.h"

//**************************************
//**                                  **
//**  Objects decribing an expression **
//**  inside an abstract tree.        **
//**                                  **
//**************************************

ostream& operator <<( ostream& stream, ExprType type )
{
  switch(type)
  {
    case Ident:
      return stream << "Identifier";
    case IntegerConst:
      return stream << "Integer Constant";
    case IntegerExpr:
      return stream << "Integer expression";
    case RealConst:
      return stream << "Real Constant";
    case RealExpr:
      return stream << "Real expression";
    case BoolConst:
      return stream << "Boolean Constant";
    case BoolExpr:
      return stream << "Boolean expression";
    case ObjExpr:
      return stream << "Object expression";
    case ObjConst:
      return stream << "Object Constant";
    case Other:
      return stream << "Other";
    case ExprError:
      return stream << "Error Type";
    default:
      return stream << "Unknown value";
  }
}

ostream& operator <<( ostream& stream, ArithOpType type )
{
  switch(type)
  {
    case Plus:
      return stream << " Plus ";
    case Minus:
      return stream << " Minus ";
    case Multiply:
      return stream << " Multiply ";
    case Divide:
      return stream << " Divide ";
    case IntDivide:
      return stream << " Integer Divide";
    case Modulo:
      return stream << " Modulo ";
    default:
      return stream << "Unknown value";
  }
}

ostream& operator <<( ostream& stream, BoolOpType type )
{
  switch(type)
  {
    case Less:
      return stream << " Less ";
    case LessOrEqual:
      return stream << " Less Or Equal ";
    case Equal:
      return stream << " Equal ";
    case NotEqual:
      return stream << " Not Equal ";
    case Greater:
      return stream << " Greater ";
    case GreaterOrEqual:
      return stream << " Greater Or Equal ";
    case Or:
      return stream << " Or ";
    case And:
      return stream << " And ";
    case AndIf:
      return stream << " AndIf ";
    case OrIf:
      return stream << " OrIf ";
    default:
      return stream << "Unknown value";
  }
}

ostream& operator <<( ostream& stream, ObjOpType type )
{
  switch(type)
  {
    case Qua:
      return stream << " Qua ";
    case Is:
      return stream << " Is ";
    case In:
      return stream << " In ";
    default:
      return stream << "Unknown value";
  }
}

Expression::Expression( ExprType rettype , Location *objloc )
{
  kind       = "NullObject";
  returntype = rettype;
  place      = objloc;
}

ostream& operator << ( ostream& stream , Expression ThisExp )
{
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream << "\nExpression :";
#endif
  ThisExp.Print( stream );
  return stream;
}

void Expression::Print( ostream& stream )
{
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream << "\nkind     :" << kind  << \
            "\ntype     :" << type  << \
            "\nlocation :" << *place << \
            "\nReturnType :" << returntype;
#endif
}

IntegerConstant::IntegerConstant( int value, Location *TheLoc ) :
       Expression( IntegerConst,TheLoc ),rvalue(value)
{
  kind   = "IntegerConstant";
}

void IntegerConstant::Print( ostream& stream )
{
  Expression::Print(stream);
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream = stream << "\nRValue  :" << rvalue;
#else
  stream << rvalue;
#endif
}

RealConstant::RealConstant( double value, Location *TheLoc ) :
       Expression( RealConst,TheLoc ),rvalue(value)
{
  kind   = "RealConstant";
}

void RealConstant::Print( ostream& stream )
{
  Expression::Print(stream);
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream = stream << "\nRValue  :" << rvalue;
#else
  stream << rvalue;
#endif
}

BoolConstant::BoolConstant( int value, Location *TheLoc ) :
       Expression( BoolConst, TheLoc ),rvalue(value)
{
  kind   = "RealConst";
}

void BoolConstant::Print( ostream& stream )
{
  Expression::Print(stream);
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream << "\nRValue  :" << rvalue;
#else
  stream << rvalue;
#endif
}

StringConstant::StringConstant( String *value, Location *TheLoc ) :
       Expression( StringConst,TheLoc ),rvalue(value)
{
  kind   = "StringConstant";
}

void StringConstant::Print( ostream& stream )
{
  Expression::Print(stream);
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream << "\nRValue  :" << *rvalue;
#else
  stream << *rvalue;
#endif
}

CharConstant::CharConstant( char value, Location *TheLoc ) :
       Expression( CharConst, TheLoc ),rvalue(value)
{
  kind   = "CharConstant";
}

void CharConstant::Print( ostream& stream )
{
  Expression::Print(stream);
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream << "\nRValue  :" << rvalue;
#else
  stream << rvalue;
#endif
}

Identifier::Identifier( String * TheName, Location *TheLoc ) :
            Expression( Ident, TheLoc )
{
  kind  = "Identifier";
  Title = TheName;
}

Identifier::Identifier( Entry ThisEntry, Location *TheLoc ) :
            Expression( Ident, TheLoc )
{
  kind  = "Identifier";
  Title = NULL;
  Where = ThisEntry;
}

void Identifier::Print( ostream& stream )
{
  Expression::Print(stream);
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream << "\nTitle :" << *Title;
#else
  stream << *Title ;
#endif
}

NoneObject::NoneObject( Location *TheLoc ) : Expression( ObjConst, TheLoc )
{
  kind = "None";
}

Result::Result( Location *TheLoc ) : Expression( ObjConst, TheLoc )
{
  kind = "Result";
}

UnaryOperator::UnaryOperator( Expression * ThisExp, ExprType ret, Location *TheLoc ) :
           Expression( ret, TheLoc )
{
  kind = "UnaryOperator";
  Argument = ThisExp;
  place = new Location(*(ThisExp->place) + *TheLoc);
}

void UnaryOperator::Print ( ostream& stream )
{
  Expression::Print(stream);
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream << "\n((";
  Argument->Print(stream);
  stream << "\n))";
#else
  stream << '(';
  Argument->Print(stream);
  stream << ')';
#endif
}

BinaryOperator::BinaryOperator( Expression *LeftExpr, Expression *RightExpr,
                                ExprType ret        , Location *TheLoc ) :
                Expression( ret, TheLoc )
{
  kind = "BinaryOperator";
  LeftMember  = LeftExpr;
  RightMember = RightExpr;
  place = new Location(*(LeftMember->place) + *TheLoc + *(RightMember->place));
}

virtual void BinaryOperator::Print( ostream& stream )
{
  Expression::Print( stream );
#ifdef VERBOSE_PRINT_EXPR_TREE
  stream << "\n(\nLeft Member :\n";
  LeftMember->Print( stream );
  stream << "  ,  ";
  RightMember->Print( stream );
  stream << "\n)";
#else
  stream << " ( ";
  LeftMember->Print(stream);
  stream << ' ' << kind << ' ';
  RightMember->Print(stream);
  stream << " ) ";
  
#endif
}

BoolOperator::BoolOperator( Expression *LeftExpr, Expression *RightExpr,
                            BoolOpType  TheOp   , Location   *TheLoc ) :
              BinaryOperator( LeftExpr, RightExpr, Other,TheLoc )
{
  kind   = "Boolean Operator";
  OpKind = TheOp;

  if ( (TheOp == Or) || (TheOp == And) )
    switch( LeftExpr->returntype)
    {
      case BoolExpr:
      case BoolConst:
      case Ident:
        switch( RightExpr->returntype )
        {
          case BoolExpr:
          case BoolConst:
          case Ident:
            returntype = BoolExpr;
            break;

          default:
            returntype = ExprError;
            cout << "Operand 2 of Operator " << TheOp << "  : \"";
            RightExpr->Print( cout );
            cout << "\" is not a boolean expression.\n";
            break;
        }
        break;

      default:
        cout << "Operand 1 of Operator " << TheOp << "  : \"";
        LeftExpr->Print( cout );
        cout << "\" is not a boolean expression.\n";
        switch( RightExpr->returntype )
        {
          case BoolExpr:
          case BoolConst:
          case Ident:
            break;

          default:
            cout << "Operand 2 of Operator " << TheOp << "  : \"";
            RightExpr->Print( cout );
            cout << "\" is not a boolan expression.\n";
            break;
        }
        returntype = ExprError;
    }
  else
    switch( LeftExpr->returntype )
    {
      case IntegerConst:
      case IntegerExpr:
        switch( RightExpr->returntype )
        {
          case IntegerConst:
          case IntegerExpr:
          case Ident:
            returntype = BoolExpr;
            break;

          default:
            returntype = ExprError;
            cout << "As operand 1 of operator " << TheOp << " is "
                 << LeftExpr->returntype << ", operand 2 that is \"";
            RightExpr->Print( cout );
            cout << "\" is expected to be an Integer Expression too.\n";
        }
        break;

      case RealConst:
      case RealExpr:
        switch( RightExpr->returntype )
        {
          case RealConst:
          case RealExpr:
          case Ident:
            returntype = BoolExpr;
            break;

          default:
            returntype = ExprError;
            cout << "As operand 1 of operator " << TheOp << " is "
                 << LeftExpr->returntype << ", operand 2 that is \"";
            RightExpr->Print( cout );
            cout << "\" is expected to be a Real Expression too.\n";
        }
        break;

      case CharConst:
      case CharExpr:
        switch( RightExpr->returntype )
        {
          case CharConst:
          case CharExpr:
          case Ident:
            returntype = BoolExpr;
            break;

          default:
            returntype = ExprError;
            cout << "As operand 1 of operator " << TheOp << " is "
                 << LeftExpr->returntype << ", operand 2 that is \"";
            RightExpr->Print( cout );
            cout << "\" is expected to be a Character Expression too.\n";
        }
        break;

      case StringConst:
      case StringExpr:
        switch( RightExpr->returntype )
        {
          case StringConst:
          case StringExpr:
          case Ident:
            returntype = BoolExpr;
            break;

          default:
            returntype = ExprError;
            cout << "As operand 1 of operator " << TheOp << " is "
                 << LeftExpr->returntype << ", operand 2 that is \"";
            RightExpr->Print( cout );
            cout << "\" is expected to be a String Expression too.\n";
        }
        break;

      case ObjConst:
      case ObjExpr:
        if ( (TheOp == Equal) || (TheOp == NotEqual) )
          switch( RightExpr->returntype )
          {
            case ObjConst:
            case ObjExpr:
            case Ident:
              returntype = BoolExpr;
              break;

            default:
              returntype = ExprError;
              cout << "As operand 1 of operator " << TheOp << " is "
                   << LeftExpr->returntype << ", operand 2 that is \"";
              RightExpr->Print( cout );
              cout << "\" is expected to be an Object Expression too.\n";
          }
        else
        {
          cout << " Only = and <> can be applied onto objects.\n";
          returntype = ExprError;
        }
        break;

      case BoolConst:
      case BoolExpr:
        if ( (TheOp == Equal) || (TheOp == NotEqual) )
          switch( RightExpr->returntype )
          {
            case BoolConst:
            case BoolExpr:
            case Ident:
              returntype = BoolExpr;
              break;

            default:
              returntype = ExprError;
              cout << "As operand 1 of operator " << TheOp << " is "
                   << LeftExpr->returntype << ", operand 2 that is \"";
              RightExpr->Print( cout );
              cout << "\" is expected to be a Boolean Expression too.\n";
          }
        else
        {
          cout << " Only = and <> can be applied onto booleans.\n";
          returntype = ExprError;
        }
        break;

      case Ident:
        switch( RightExpr->returntype )
        {
          case IntegerConst:
          case IntegerExpr:
          case RealConst:
          case RealExpr:
          case CharConst:
          case CharExpr:
          case StringConst:
          case StringExpr:
            returntype = BoolExpr;
            break;

          case ObjConst:
          case ObjExpr:
            if ( (TheOp == Equal) || (TheOp == NotEqual) )
              returntype = BoolExpr;
            else
            {
              cout << " You can only apply Equal and NotEqual onto objects.\n";
              returntype = ExprError;
            }
            break;

          case BoolConst:
          case BoolExpr:
            if ( (TheOp == Equal) || (TheOp == NotEqual) )
              returntype = BoolExpr;
            else
            {
              cout << " You can only apply Equal and NotEqual onto booleans.\n";
              returntype = ExprError;
            }
            break;

          case Ident:
            returntype = BoolExpr;
            break;

          default:
            returntype = ExprError;
            cout << " Type of Operand 2 \"";
            RightExpr->Print( cout );
            cout << "\" of operand " << TheOp << " is wrong.\n";
        }
        break;

      case ExprError:
        returntype = ExprError;
        break;

      default:
        returntype = ExprError;
        cout << " The type of Operand 1 \"";
        LeftExpr->Print( cout );
        cout << "\" of operator " << TheOp << " is wrong.\n";
    }
}

virtual void BoolOperator::Print( ostream& stream )
{
#ifdef VERBOSE_PRINT_EXPR_TREE
  BinaryOperator::Print( stream );
  stream << "\nKind of Boolean Operator : " << OpKind;
#else
  stream << " ( ";
  LeftMember->Print( stream );
  stream << ' ' << OpKind << ' ';
  RightMember->Print( stream );
  stream << " ) ";
#endif
}

ArithOperator::ArithOperator( Expression  *LeftExpr, Expression *RightExpr,
                              ArithOpType  TheOp   , Location   *TheLoc ) :
               BinaryOperator( LeftExpr, RightExpr, Other, TheLoc )
{
  kind   = "Arithmetic Operator";
  OpKind = TheOp;

  switch( LeftExpr->returntype)
  {
    case IntegerConst:
    case IntegerExpr:

      switch( RightExpr->returntype )
      {
        case IntegerConst:
        case IntegerExpr:
          returntype = IntegerExpr;
          break;

        case RealConst:
        case RealExpr:
          returntype = RealExpr;
          break;

        case Ident:
          returntype = IntegerExpr;
          break;

        case ExprError:
          returntype = ExprError;
          break;

        default:
          returntype = ExprError;
          cout << "Operand 2 of Operator " << TheOp << "  : \"";
          RightExpr->Print( cout );
          cout << "\" is not an arithmetic expression.\n";
          break;
      }
      break;

    case RealConst:
    case RealExpr:
      switch( RightExpr->returntype )
      {

        case IntegerConst:
        case IntegerExpr:
        case RealConst:
        case RealExpr:
        case Ident:
          returntype = RealExpr;
          break;

        case ExprError:
          returntype = ExprError;
          break;

        default:
          returntype = ExprError;
          cout << "Operand 2 of Operator " << TheOp << "  : \"";
          RightExpr->Print( cout );
          cout << "\" is not an arithmetic expression.\n";

      }
      break;

    case Ident:
      switch( RightExpr->returntype )
      {
        case IntegerConst:
        case IntegerExpr:
          returntype = IntegerExpr;
          break;

        case RealConst:
        case RealExpr:
          returntype = RealExpr;
          break;

        case Ident:
          returntype = Ident;
          break;

        case ExprError:
          returntype = ExprError;
          break;

        default:
          returntype = ExprError;
          cout << "Operand 2 of Operator " << TheOp << "  : \"";
          RightExpr->Print( cout );
          cout << "\" is not an arithmetic expression.\n";
      }
      break;

    case ExprError:
      returntype = ExprError;
      break;

    default:
      returntype = ExprError;
      cout << "Operand 1 of Operator " << TheOp << "  : \"";
      LeftExpr->Print( cout );
      cout << "\" is not an arithmetic expression.\n";
      switch( RightExpr->returntype )
      {
        case IntegerConst:
        case IntegerExpr:
        case RealConst:
        case RealExpr:
        case Ident:
          break;

        default:
          cout << "Operand 2 of Operator " << TheOp << "  : \"";
          RightExpr->Print( cout );
          cout << "\" is not an arithmetic expression.\n";
      }

  }
}

virtual void ArithOperator::Print( ostream& stream)
{
#ifdef VERBOSE_PRINT_EXPR_TREE
  BinaryOperator::Print( stream );
  stream << "\nKind of Arithmetic Operator : " << OpKind;
#else
  stream << " ( ";
  LeftMember->Print( stream );
  stream << ' ' << OpKind << ' ';
  RightMember->Print( stream );
  stream << " ) ";
#endif
}

ObjOperator::ObjOperator( Expression *LeftExpr, Expression  *RightExpr,
                          ObjOpType   TheOp   , Location    *TheLoc ) :
              BinaryOperator( LeftExpr, RightExpr, Other, TheLoc )
{
  kind   = "Object Operator";
  OpKind = TheOp;

  switch(LeftExpr->returntype)
  {
    case Ident:
      switch ( RightExpr->returntype )
      {
        case Ident:
          if (TheOp == Qua)
	    returntype = ObjExpr;
          else
            returntype = BoolExpr;
          break;

        case ExprError:
	  returntype = ExprError;
          break;

        default:
	  returntype = ExprError;
          cout << "Operand 2 of Operator " << TheOp << "  : \"";
          RightExpr->Print( cout );
          cout << "\" is not an Identifier.\n";
      }
      break;

    case ExprError:
      returntype = ExprError;
      break;

    default:
      cout << "Operand 1 of Operator " << TheOp << "  : \"";
      LeftExpr->Print( cout );
      cout << "\" is not an Identifier.\n";

  }
}

virtual void ObjOperator::Print( ostream& stream )
{
#ifdef VERBOSE_PRINT_EXPR_TREE
  BinaryOperator::Print( stream );
  stream << "\nKind of Object Operator : " << OpKind;
#else
  stream << " ( ";
  LeftMember->Print( stream );
  stream << ' ' << OpKind << ' ';
  RightMember->Print( stream );
  stream << " ) ";
#endif
}

Not::Not( Expression *TheExpr, Location *TheLoc ) :
     UnaryOperator( TheExpr , Other, TheLoc )
{
  kind = "Not";

  switch( TheExpr->returntype )
  {
    case BoolExpr:
    case BoolConst:
    case Ident:
      returntype = BoolExpr;
      break;

    case ExprError:
      returntype = ExprError;
      break;

    default:
      returntype = ExprError;
      cout << "Operand of Operator Not  : \"";
      TheExpr->Print( cout );
      cout << "\" is not a Boolean Expression.\n";
  }
}

This::This( Expression *TheExpr, Location *TheLoc ) :
      UnaryOperator( TheExpr, Other, TheLoc )
{
  kind = "This";

  if ( TheExpr )
    switch ( TheExpr->returntype )
    {
      case Ident:
        returntype = ObjExpr;
	break;

      default:
        returntype = ExprError;
	cout << "Operand is of operator 'This' is not an Identifier.\n";
    }
  else
    returntype = ObjExpr;
}

void This::Print( ostream& stream )
{
  cout << " This ";
  if ( Argument )
    UnaryOperator::Print( stream );
}

New::New( Expression *TheExpr, Location *TheLoc ) :
     UnaryOperator( TheExpr , Other, TheLoc )
{
  kind = "New";

  switch( TheExpr->returntype )
  {
    case Ident:
      returntype = ObjExpr;
      break;

    case ExprError:
      returntype = ExprError;
      break;

    default:
      returntype = ExprError;
      cout << "Operand \"";
      TheExpr->Print(cout);
      cout << "\"type of New operator is wrong.\n";
  }
}

Copy::Copy( Expression *TheExpr, Location *TheLoc ) :
      UnaryOperator( TheExpr, Other, TheLoc )
{
  kind = "Copy";
  switch(TheExpr->returntype)
  {
    case Ident:
    case ObjExpr:
      returntype = ObjExpr;
      break;

    case ExprError:
      returntype = ExprError;
      break;

    default:
      returntype = ExprError;
      cout << "Wrong operand type for operand \"";
      TheExpr->Print(cout);
      cout << "\" of Copy.\n";
      break;
  }
}

ArrayOf::ArrayOf( Expression *TheExpr, ExprType ret, Location *TheLoc ) :
         UnaryOperator( TheExpr, ret, TheLoc )
{
  kind = "ArrayOf";
}

Error::Error( ) : Expression( ExprError, NULL )
{
  kind = "Error into Expression";
}
